from __future__ import annotations

from dataclasses import dataclass, replace

from .. import random_id
from ..style import Style, NamedStyles


class PythonStyle(Style):
    pass


@dataclass(frozen=True)
class SpacesStyle(PythonStyle):
    @dataclass(frozen=True)
    class BeforeParentheses:
        _method_call: bool

        @property
        def method_call(self) -> bool:
            return self._method_call

        def with_method_call(self, method_call: bool) -> SpacesStyle.BeforeParentheses:
            return self if method_call is self._method_call else replace(self, _method_call=method_call)

        _method_declaration: bool

        @property
        def method_declaration(self) -> bool:
            return self._method_declaration

        def with_method_declaration(self, method_declaration: bool) -> SpacesStyle.BeforeParentheses:
            return self if method_declaration is self._method_declaration else replace(self, _method_declaration=method_declaration)

        _left_bracket: bool

        @property
        def left_bracket(self) -> bool:
            return self._left_bracket

        def with_left_bracket(self, left_bracket: bool) -> SpacesStyle.BeforeParentheses:
            return self if left_bracket is self._left_bracket else replace(self, _left_bracket=left_bracket)

    @dataclass(frozen=True)
    class AroundOperators:
        _assignment: bool

        @property
        def assignment(self) -> bool:
            return self._assignment

        def with_assignment(self, assignment: bool) -> SpacesStyle.AroundOperators:
            return self if assignment is self._assignment else replace(self, _assignment=assignment)

        _equality: bool

        @property
        def equality(self) -> bool:
            return self._equality

        def with_equality(self, equality: bool) -> SpacesStyle.AroundOperators:
            return self if equality is self._equality else replace(self, _equality=equality)

        _relational: bool

        @property
        def relational(self) -> bool:
            return self._relational

        def with_relational(self, relational: bool) -> SpacesStyle.AroundOperators:
            return self if relational is self._relational else replace(self, _relational=relational)

        _bitwise: bool

        @property
        def bitwise(self) -> bool:
            return self._bitwise

        def with_bitwise(self, bitwise: bool) -> SpacesStyle.AroundOperators:
            return self if bitwise is self._bitwise else replace(self, _bitwise=bitwise)

        _additive: bool

        @property
        def additive(self) -> bool:
            return self._additive

        def with_additive(self, additive: bool) -> SpacesStyle.AroundOperators:
            return self if additive is self._additive else replace(self, _additive=additive)

        _multiplicative: bool

        @property
        def multiplicative(self) -> bool:
            return self._multiplicative

        def with_multiplicative(self, multiplicative: bool) -> SpacesStyle.AroundOperators:
            return self if multiplicative is self._multiplicative else replace(self, _multiplicative=multiplicative)

        _shift: bool

        @property
        def shift(self) -> bool:
            return self._shift

        def with_shift(self, shift: bool) -> SpacesStyle.AroundOperators:
            return self if shift is self._shift else replace(self, _shift=shift)

        _power: bool

        @property
        def power(self) -> bool:
            return self._power

        def with_power(self, power: bool) -> SpacesStyle.AroundOperators:
            return self if power is self._power else replace(self, _power=power)

        _eq_in_named_parameter: bool

        @property
        def eq_in_named_parameter(self) -> bool:
            return self._eq_in_named_parameter

        def with_eq_in_named_parameter(self, eq_in_named_parameter: bool) -> SpacesStyle.AroundOperators:
            return self if eq_in_named_parameter is self._eq_in_named_parameter else replace(self, _eq_in_named_parameter=eq_in_named_parameter)

        _eq_in_keyword_argument: bool

        @property
        def eq_in_keyword_argument(self) -> bool:
            return self._eq_in_keyword_argument

        def with_eq_in_keyword_argument(self, eq_in_keyword_argument: bool) -> SpacesStyle.AroundOperators:
            return self if eq_in_keyword_argument is self._eq_in_keyword_argument else replace(self, _eq_in_keyword_argument=eq_in_keyword_argument)

    @dataclass(frozen=True)
    class Within:
        _brackets: bool

        @property
        def brackets(self) -> bool:
            return self._brackets

        def with_brackets(self, brackets: bool) -> SpacesStyle.Within:
            return self if brackets is self._brackets else replace(self, _brackets=brackets)

        _method_declaration_parentheses: bool

        @property
        def method_declaration_parentheses(self) -> bool:
            return self._method_declaration_parentheses

        def with_method_declaration_parentheses(self, method_declaration_parentheses: bool) -> SpacesStyle.Within:
            return self if method_declaration_parentheses is self._method_declaration_parentheses else replace(self, _method_declaration_parentheses=method_declaration_parentheses)

        _empty_method_declaration_parentheses: bool

        @property
        def empty_method_declaration_parentheses(self) -> bool:
            return self._empty_method_declaration_parentheses

        def with_empty_method_declaration_parentheses(self, empty_method_declaration_parentheses: bool) -> SpacesStyle.Within:
            return self if empty_method_declaration_parentheses is self._empty_method_declaration_parentheses else replace(self, _empty_method_declaration_parentheses=empty_method_declaration_parentheses)

        _method_call_parentheses: bool

        @property
        def method_call_parentheses(self) -> bool:
            return self._method_call_parentheses

        def with_method_call_parentheses(self, method_call_parentheses: bool) -> SpacesStyle.Within:
            return self if method_call_parentheses is self._method_call_parentheses else replace(self, _method_call_parentheses=method_call_parentheses)

        _empty_method_call_parentheses: bool

        @property
        def empty_method_call_parentheses(self) -> bool:
            return self._empty_method_call_parentheses

        def with_empty_method_call_parentheses(self, empty_method_call_parentheses: bool) -> SpacesStyle.Within:
            return self if empty_method_call_parentheses is self._empty_method_call_parentheses else replace(self, _empty_method_call_parentheses=empty_method_call_parentheses)

        _braces: bool

        @property
        def braces(self) -> bool:
            return self._braces

        def with_braces(self, braces: bool) -> SpacesStyle.Within:
            return self if braces is self._braces else replace(self, _braces=braces)

    @dataclass(frozen=True)
    class Other:
        _before_comma: bool

        @property
        def before_comma(self) -> bool:
            return self._before_comma

        def with_before_comma(self, before_comma: bool) -> SpacesStyle.Other:
            return self if before_comma is self._before_comma else replace(self, _before_comma=before_comma)

        _after_comma: bool

        @property
        def after_comma(self) -> bool:
            return self._after_comma

        def with_after_comma(self, after_comma: bool) -> SpacesStyle.Other:
            return self if after_comma is self._after_comma else replace(self, _after_comma=after_comma)

        _before_for_semicolon: bool

        @property
        def before_for_semicolon(self) -> bool:
            return self._before_for_semicolon

        def with_before_for_semicolon(self, before_for_semicolon: bool) -> SpacesStyle.Other:
            return self if before_for_semicolon is self._before_for_semicolon else replace(self, _before_for_semicolon=before_for_semicolon)

        _before_colon: bool

        @property
        def before_colon(self) -> bool:
            return self._before_colon

        def with_before_colon(self, before_colon: bool) -> SpacesStyle.Other:
            return self if before_colon is self._before_colon else replace(self, _before_colon=before_colon)

        _after_colon: bool

        @property
        def after_colon(self) -> bool:
            return self._after_colon

        def with_after_colon(self, after_colon: bool) -> SpacesStyle.Other:
            return self if after_colon is self._after_colon else replace(self, _after_colon=after_colon)

        _before_backslash: bool

        @property
        def before_backslash(self) -> bool:
            return self._before_backslash

        def with_before_backslash(self, before_backslash: bool) -> SpacesStyle.Other:
            return self if before_backslash is self._before_backslash else replace(self, _before_backslash=before_backslash)

        _before_hash: bool

        @property
        def before_hash(self) -> bool:
            return self._before_hash

        def with_before_hash(self, before_hash: bool) -> SpacesStyle.Other:
            return self if before_hash is self._before_hash else replace(self, _before_hash=before_hash)

        _after_hash: bool

        @property
        def after_hash(self) -> bool:
            return self._after_hash

        def with_after_hash(self, after_hash: bool) -> SpacesStyle.Other:
            return self if after_hash is self._after_hash else replace(self, _after_hash=after_hash)

    _before_parentheses: BeforeParentheses

    @property
    def before_parentheses(self) -> BeforeParentheses:
        return self._before_parentheses

    def with_before_parentheses(self, before_parentheses: BeforeParentheses) -> SpacesStyle:
        return self if before_parentheses is self._before_parentheses else replace(self, _before_parentheses=before_parentheses)

    _around_operators: AroundOperators

    @property
    def around_operators(self) -> AroundOperators:
        return self._around_operators

    def with_around_operators(self, around_operators: AroundOperators) -> SpacesStyle:
        return self if around_operators is self._around_operators else replace(self, _around_operators=around_operators)

    _within: Within

    @property
    def within(self) -> Within:
        return self._within

    def with_within(self, within: Within) -> SpacesStyle:
        return self if within is self._within else replace(self, _within=within)

    _other: Other

    @property
    def other(self) -> Other:
        return self._other

    def with_other(self, other: Other) -> SpacesStyle:
        return self if other is self._other else replace(self, _other=other)


@dataclass(frozen=True)
class TabsAndIndentsStyle(PythonStyle):
    _use_tab_character: bool

    @property
    def use_tab_character(self) -> bool:
        return self._use_tab_character

    def with_use_tab_character(self, use_tab_character: bool) -> TabsAndIndentsStyle:
        return self if use_tab_character is self._use_tab_character else replace(self, _use_tab_character=use_tab_character)

    _tab_size: int

    @property
    def tab_size(self) -> int:
        return self._tab_size

    def with_tab_size(self, tab_size: int) -> TabsAndIndentsStyle:
        return self if tab_size is self._tab_size else replace(self, _tab_size=tab_size)

    _indent_size: int

    @property
    def indent_size(self) -> int:
        return self._indent_size

    def with_indent_size(self, indent_size: int) -> TabsAndIndentsStyle:
        return self if indent_size is self._indent_size else replace(self, _indent_size=indent_size)

    _continuation_indent: int

    @property
    def continuation_indent(self) -> int:
        return self._continuation_indent

    def with_continuation_indent(self, continuation_indent: int) -> TabsAndIndentsStyle:
        return self if continuation_indent is self._continuation_indent else replace(self, _continuation_indent=continuation_indent)

    _keep_indents_on_empty_lines: bool

    @property
    def keep_indents_on_empty_lines(self) -> bool:
        return self._keep_indents_on_empty_lines

    def with_keep_indents_on_empty_lines(self, keep_indents_on_empty_lines: bool) -> TabsAndIndentsStyle:
        return self if keep_indents_on_empty_lines is self._keep_indents_on_empty_lines else replace(self, _keep_indents_on_empty_lines=keep_indents_on_empty_lines)

    @dataclass(frozen=True)
    class MethodDeclarationParameters:
        _align_multiline_parameters: bool

        @property
        def align_multiline_parameters(self) -> bool:
            return self._align_multiline_parameters

        def with_align_multiline_parameters(self,
                                            align_multiline_parameters: bool) -> TabsAndIndentsStyle.MethodDeclarationParameters:
            return self if align_multiline_parameters is self._align_multiline_parameters else replace(self,
                                                                                                       _align_multiline_parameters=align_multiline_parameters)

    _method_declaration_parameters: MethodDeclarationParameters

    @property
    def method_declaration_parameters(self) -> MethodDeclarationParameters:
        return self._method_declaration_parameters

    def with_method_declaration_parameters(self,
                                           method_declaration_parameters: MethodDeclarationParameters) -> TabsAndIndentsStyle:
        return self if method_declaration_parameters is self._method_declaration_parameters else replace(self,
                                                                                                         _method_declaration_parameters=method_declaration_parameters)


@dataclass(frozen=True)
class WrappingAndBracesStyle(PythonStyle):
    pass


@dataclass(frozen=True)
class BlankLinesStyle(PythonStyle):
    @dataclass(frozen=True)
    class KeepMaximum:
        _in_declarations: int

        @property
        def in_declarations(self) -> int:
            return self._in_declarations

        def with_in_declarations(self, in_declarations: int) -> BlankLinesStyle.KeepMaximum:
            return self if in_declarations is self._in_declarations else replace(self, _in_declarations=in_declarations)

        _in_code: int

        @property
        def in_code(self) -> int:
            return self._in_code

        def with_in_code(self, in_code: int) -> BlankLinesStyle.KeepMaximum:
            return self if in_code is self._in_code else replace(self, _in_code=in_code)

    @dataclass(frozen=True)
    class Minimum:
        _after_top_level_imports: int

        @property
        def after_top_level_imports(self) -> int:
            return self._after_top_level_imports

        def with_after_top_level_imports(self, after_top_level_imports: int) -> BlankLinesStyle.Minimum:
            return self if after_top_level_imports is self._after_top_level_imports else replace(self, _after_top_level_imports=after_top_level_imports)

        _around_class: int

        @property
        def around_class(self) -> int:
            return self._around_class

        def with_around_class(self, around_class: int) -> BlankLinesStyle.Minimum:
            return self if around_class is self._around_class else replace(self, _around_class=around_class)

        _around_method: int

        @property
        def around_method(self) -> int:
            return self._around_method

        def with_around_method(self, around_method: int) -> BlankLinesStyle.Minimum:
            return self if around_method is self._around_method else replace(self, _around_method=around_method)

        _around_top_level_classes_functions: int

        @property
        def around_top_level_classes_functions(self) -> int:
            return self._around_top_level_classes_functions

        def with_around_top_level_classes_functions(self, around_top_level_classes_functions: int) -> BlankLinesStyle.Minimum:
            return self if around_top_level_classes_functions is self._around_top_level_classes_functions else replace(self, _around_top_level_classes_functions=around_top_level_classes_functions)

        _after_local_imports: int

        @property
        def after_local_imports(self) -> int:
            return self._after_local_imports

        def with_after_local_imports(self, after_local_imports: int) -> BlankLinesStyle.Minimum:
            return self if after_local_imports is self._after_local_imports else replace(self, _after_local_imports=after_local_imports)

        _before_first_method: int

        @property
        def before_first_method(self) -> int:
            return self._before_first_method

        def with_before_first_method(self, before_first_method: int) -> BlankLinesStyle.Minimum:
            return self if before_first_method is self._before_first_method else replace(self, _before_first_method=before_first_method)

    _keep_maximum: KeepMaximum

    @property
    def keep_maximum(self) -> KeepMaximum:
        return self._keep_maximum

    def with_keep_maximum(self, keep_maximum: KeepMaximum) -> BlankLinesStyle:
        return self if keep_maximum is self._keep_maximum else replace(self, _keep_maximum=keep_maximum)

    _minimum: Minimum

    @property
    def minimum(self) -> Minimum:
        return self._minimum

    def with_minimum(self, minimum: Minimum) -> BlankLinesStyle:
        return self if minimum is self._minimum else replace(self, _minimum=minimum)


@dataclass(frozen=True)
class OtherStyle(PythonStyle):
    @dataclass(frozen=True)
    class UseContinuationIndent:
        _method_call_arguments: bool

        @property
        def method_call_arguments(self) -> bool:
            return self._method_call_arguments

        def with_method_call_arguments(self, method_call_arguments: bool) -> OtherStyle.UseContinuationIndent:
            return self if method_call_arguments is self._method_call_arguments else replace(self, _method_call_arguments=method_call_arguments)

        _method_declaration_parameters: bool

        @property
        def method_declaration_parameters(self) -> bool:
            return self._method_declaration_parameters

        def with_method_declaration_parameters(self, method_declaration_parameters: bool) -> OtherStyle.UseContinuationIndent:
            return self if method_declaration_parameters is self._method_declaration_parameters else replace(self, _method_declaration_parameters=method_declaration_parameters)

        _collections_and_comprehensions: bool

        @property
        def collections_and_comprehensions(self) -> bool:
            return self._collections_and_comprehensions

        def with_collections_and_comprehensions(self, collections_and_comprehensions: bool) -> OtherStyle.UseContinuationIndent:
            return self if collections_and_comprehensions is self._collections_and_comprehensions else replace(self, _collections_and_comprehensions=collections_and_comprehensions)

    _use_continuation_indent: UseContinuationIndent

    @property
    def use_continuation_indent(self) -> UseContinuationIndent:
        return self._use_continuation_indent

    def with_use_continuation_indent(self, use_continuation_indent: UseContinuationIndent) -> OtherStyle:
        return self if use_continuation_indent is self._use_continuation_indent else replace(self, _use_continuation_indent=use_continuation_indent)


class IntelliJ(NamedStyles):

    def __init__(self):
        super().__init__(
            _id=random_id(),
            _name='org.openrewrite.python.style.IntelliJ',
            _display_name='IntelliJ IDEA',
            _description='IntelliJ IDEA default Python style.',
            _tags={},
            _styles=(
                IntelliJ.spaces(),
                IntelliJ.wrapping_and_braces(),
                IntelliJ.tabs_and_indents(),
                IntelliJ.blank_lines(),
                IntelliJ.other(),
            )
        )

    @classmethod
    def defaults(cls) -> IntelliJ:
        return IntelliJ()

    @classmethod
    def spaces(cls) -> SpacesStyle:
        return SpacesStyle(
            SpacesStyle.BeforeParentheses(
                _method_call=False,
                _method_declaration=False,
                _left_bracket=False,
            ),
            SpacesStyle.AroundOperators(
                _assignment=True,
                _equality=True,
                _relational=True,
                _bitwise=True,
                _additive=True,
                _multiplicative=True,
                _shift=True,
                _power=True,
                _eq_in_named_parameter=False,
                _eq_in_keyword_argument=False,
            ),
            SpacesStyle.Within(
                _brackets=False,
                _method_declaration_parentheses=False,
                _empty_method_declaration_parentheses=False,
                _method_call_parentheses=False,
                _empty_method_call_parentheses=False,
                _braces=False,
            ),
            SpacesStyle.Other(
                _before_comma=False,
                _after_comma=True,
                _before_for_semicolon=False,
                _before_colon=False,
                _after_colon=True,
                _before_backslash=True,
                _before_hash=True,
                _after_hash=True,
            ),
        )

    @classmethod
    def tabs_and_indents(cls) -> TabsAndIndentsStyle:
        return TabsAndIndentsStyle(
            _use_tab_character=False,
            _tab_size=4,
            _indent_size=4,
            _continuation_indent=8,
            _keep_indents_on_empty_lines=False,
            _method_declaration_parameters=TabsAndIndentsStyle.MethodDeclarationParameters(
                _align_multiline_parameters=True,
            ),
        )

    @classmethod
    def wrapping_and_braces(cls) -> WrappingAndBracesStyle:
        return WrappingAndBracesStyle()

    @classmethod
    def blank_lines(cls) -> BlankLinesStyle:
        return BlankLinesStyle(
            BlankLinesStyle.KeepMaximum(
                _in_declarations=1,
                _in_code=1,
            ),
            BlankLinesStyle.Minimum(
                _after_top_level_imports=1,
                _around_class=1,
                _around_method=1,
                _around_top_level_classes_functions=2,
                _after_local_imports=0,
                _before_first_method=0,
            )
        )

    @classmethod
    def other(cls) -> OtherStyle:
        return OtherStyle(
            OtherStyle.UseContinuationIndent(
                _method_call_arguments=False,
                _method_declaration_parameters=True,
                _collections_and_comprehensions=False,
            )
        )



# Copyright 2025 the original author or authors.
# <p>
# Licensed under the Moderne Source Available License (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# <p>
# https://docs.moderne.io/licensing/moderne-source-available-license
# <p>
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Source specification types and helpers for the Python test harness."""

from __future__ import annotations

import textwrap
from dataclasses import dataclass
from pathlib import Path
from typing import Callable, Optional, Union, TYPE_CHECKING

if TYPE_CHECKING:
    from rewrite.tree import SourceFile

# Type alias for the "after" field:
# - None: No change expected (recipe should not modify the file)
# - str: Expected source after transformation
# - Callable[[str], str | None]: Function that receives actual output and returns expected
#   (returning None means the actual output is acceptable as-is)
AfterRecipeText = Union[str, Callable[[str], Optional[str]], None]


@dataclass
class SourceSpec:
    """
    Specification for a source file in a rewrite test.

    Defines the before state, expected after state, and optional hooks
    for a single source file being tested.

    Attributes:
        kind: Language identifier (e.g., "python")
        before: Source code before the recipe runs. None for generated files.
        after: Expected source after transformation:
            - None: No change expected
            - str: Expected output
            - Callable: Function receiving actual output, returning expected
        path: Optional file path (auto-generated if not provided)
        before_recipe: Hook called after parsing but before recipe runs
        after_recipe: Hook called after recipe runs
        ext: File extension for auto-generated paths
    """

    kind: str
    before: Optional[str]
    after: AfterRecipeText = None
    path: Optional[Path] = None
    before_recipe: Optional[Callable[["SourceFile"], Optional["SourceFile"]]] = None
    after_recipe: Optional[Callable[["SourceFile"], None]] = None
    ext: str = "py"


def dedent(s: Optional[str]) -> str:
    """
    Remove common leading whitespace from a string.

    Behavior:
    - Removes ONE leading newline if present (for template string ergonomics)
    - Preserves trailing newlines
    - Removes common indentation from all lines

    This allows tests to use cleanly indented triple-quoted strings:

        python(
            '''
            def foo():
                pass
            '''
        )

    Instead of awkward formatting:

        python(
            '''def foo():
        pass
'''
        )

    Args:
        s: The string to dedent

    Returns:
        The dedented string, or empty string if input is None/empty
    """
    if not s:
        return ""

    # Remove single leading newline for ergonomics
    if s.startswith("\n"):
        s = s[1:]

    if not s:
        return ""

    # Preserve \r\n sequences through dedent: replace \r\n with a marker
    # that keeps the \n (so line boundaries survive for textwrap.dedent)
    _CR = '\x00CR\x00'
    s = s.replace('\r\n', _CR + '\n')
    s = textwrap.dedent(s)
    return s.replace(_CR + '\n', '\r\n')


def python(
    before: Optional[str],
    after: AfterRecipeText = None,
    *,
    path: Optional[Path] = None,
    before_recipe: Optional[Callable[["SourceFile"], Optional["SourceFile"]]] = None,
    after_recipe: Optional[Callable[["SourceFile"], None]] = None,
) -> SourceSpec:
    """
    Create a SourceSpec for Python source code.

    This is the primary helper function for creating test specifications
    for Python source files.

    Args:
        before: Source code before transformation (None for generated files)
        after: Expected source after transformation:
            - None: No change expected
            - str: Expected output
            - Callable: Function receiving actual, returning expected
        path: Optional file path (auto-generated if not provided)
        before_recipe: Hook called after parsing, before recipe runs.
            Can modify the parsed AST by returning a new SourceFile.
        after_recipe: Hook called after recipe runs to validate results.

    Returns:
        A SourceSpec configured for Python

    Example:
        # Test a recipe that removes unused imports
        python(
            '''
            import os
            import sys
            print("hello")
            ''',
            '''
            import sys
            print("hello")
            '''
        )

        # Test that no changes are made
        python(
            '''
            import sys
            '''
            # No after = no change expected
        )

        # Use a callable for dynamic assertions
        python(
            '''
            x = 1
            ''',
            lambda actual: actual if "x" in actual else None
        )
    """
    return SourceSpec(
        kind="python",
        before=before,
        after=after,
        path=path,
        before_recipe=before_recipe,
        after_recipe=after_recipe,
        ext="py",
    )



# Copyright 2025 the original author or authors.
# <p>
# Licensed under the Moderne Source Available License (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# <p>
# https://docs.moderne.io/licensing/moderne-source-available-license
# <p>
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for data table support."""

import os
import tempfile
from dataclasses import dataclass, field

import pytest

from rewrite import InMemoryExecutionContext
from rewrite.data_table import (
    DataTable,
    column,
    InMemoryDataTableStore,
    CsvDataTableStore,
    DATA_TABLE_STORE,
)


@dataclass
class SampleRow:
    """A test row type for data tables."""

    source_path: str = field(metadata=column("Source Path", "The path of the file"))
    text: str = field(metadata=column("Text", "Some text content"))


class TestDataTable:
    """Tests for DataTable class."""

    def test_descriptor_has_columns(self):
        """Test that descriptor includes column metadata."""
        table = DataTable[SampleRow](
            "org.openrewrite.test.TestTable",
            "Test Table",
            "A test data table.",
            SampleRow,
        )

        descriptor = table.descriptor()

        assert descriptor["name"] == "org.openrewrite.test.TestTable"
        assert descriptor["displayName"] == "Test Table"
        assert descriptor["description"] == "A test data table."
        assert len(descriptor["columns"]) == 2

        col1 = descriptor["columns"][0]
        assert col1["name"] == "source_path"
        assert col1["displayName"] == "Source Path"
        assert col1["description"] == "The path of the file"

        col2 = descriptor["columns"][1]
        assert col2["name"] == "text"
        assert col2["displayName"] == "Text"
        assert col2["description"] == "Some text content"

    def test_insert_row_creates_store(self):
        """Test that inserting a row creates an InMemoryDataTableStore if none exists."""
        table = DataTable[SampleRow](
            "org.openrewrite.test.TestTable",
            "Test Table",
            "A test data table.",
            SampleRow,
        )
        ctx = InMemoryExecutionContext()

        # Initially no store
        assert ctx.get_message(DATA_TABLE_STORE) is None

        # Insert a row
        table.insert_row(ctx, SampleRow("src/foo.py", "hello"))

        # Now there should be a store
        store = ctx.get_message(DATA_TABLE_STORE)
        assert store is not None
        assert isinstance(store, InMemoryDataTableStore)


class TestInMemoryDataTableStore:
    """Tests for InMemoryDataTableStore."""

    def test_stores_rows_when_accepting(self):
        """Test that rows are stored when accept_rows is True."""
        store = InMemoryDataTableStore()
        table = DataTable[SampleRow](
            "org.openrewrite.test.TestTable",
            "Test Table",
            "A test data table.",
            SampleRow,
        )
        ctx = InMemoryExecutionContext()

        # By default, accept_rows is False
        store.insert_row(table, ctx, SampleRow("src/foo.py", "hello"))
        assert len(store.rows) == 0

        # Enable accepting rows
        store.accept_rows(True)
        store.insert_row(table, ctx, SampleRow("src/foo.py", "hello"))
        store.insert_row(table, ctx, SampleRow("src/bar.py", "world"))

        assert "org.openrewrite.test.TestTable" in store.rows
        assert len(store.rows["org.openrewrite.test.TestTable"]) == 2


class TestCsvDataTableStore:
    """Tests for CsvDataTableStore."""

    def test_writes_csv_with_header(self):
        """Test that CSV files are written with header and data rows."""
        with tempfile.TemporaryDirectory() as tmpdir:
            store = CsvDataTableStore(tmpdir)
            table = DataTable[SampleRow](
                "org.openrewrite.test.TestTable",
                "Test Table",
                "A test data table.",
                SampleRow,
            )
            ctx = InMemoryExecutionContext()

            store.accept_rows(True)
            store.insert_row(table, ctx, SampleRow("src/foo.py", "hello"))
            store.insert_row(table, ctx, SampleRow("src/bar.py", "world"))

            csv_path = os.path.join(tmpdir, "org.openrewrite.test.TestTable.csv")
            assert os.path.exists(csv_path)

            with open(csv_path, "r") as f:
                content = f.read()

            lines = content.strip().split("\n")
            assert len(lines) == 3  # header + 2 data rows
            assert lines[0] == "Source Path,Text"
            assert lines[1] == "src/foo.py,hello"
            assert lines[2] == "src/bar.py,world"

    def test_escapes_special_characters(self):
        """Test that special characters are properly escaped in CSV."""
        with tempfile.TemporaryDirectory() as tmpdir:
            store = CsvDataTableStore(tmpdir)
            table = DataTable[SampleRow](
                "org.openrewrite.test.TestTable",
                "Test Table",
                "A test data table.",
                SampleRow,
            )
            ctx = InMemoryExecutionContext()

            store.accept_rows(True)
            # Test comma, quote, and newline escaping
            store.insert_row(table, ctx, SampleRow("src/foo.py", "hello, world"))
            store.insert_row(table, ctx, SampleRow("src/bar.py", 'say "hi"'))
            store.insert_row(table, ctx, SampleRow("src/baz.py", "line1\nline2"))

            csv_path = os.path.join(tmpdir, "org.openrewrite.test.TestTable.csv")
            with open(csv_path, "r") as f:
                content = f.read()

            lines = content.strip().split("\n")
            # Note: newlines in values will cause extra lines
            assert 'src/foo.py,"hello, world"' in content
            assert 'src/bar.py,"say ""hi"""' in content

    def test_does_not_write_when_not_accepting(self):
        """Test that no rows are written when accept_rows is False."""
        with tempfile.TemporaryDirectory() as tmpdir:
            store = CsvDataTableStore(tmpdir)
            table = DataTable[SampleRow](
                "org.openrewrite.test.TestTable",
                "Test Table",
                "A test data table.",
                SampleRow,
            )
            ctx = InMemoryExecutionContext()

            # Don't enable accepting rows
            store.insert_row(table, ctx, SampleRow("src/foo.py", "hello"))

            csv_path = os.path.join(tmpdir, "org.openrewrite.test.TestTable.csv")
            assert not os.path.exists(csv_path)

    def test_tracks_row_counts(self):
        """Test that row counts are tracked correctly."""
        with tempfile.TemporaryDirectory() as tmpdir:
            store = CsvDataTableStore(tmpdir)
            table = DataTable[SampleRow](
                "org.openrewrite.test.TestTable",
                "Test Table",
                "A test data table.",
                SampleRow,
            )
            ctx = InMemoryExecutionContext()

            store.accept_rows(True)
            store.insert_row(table, ctx, SampleRow("src/foo.py", "hello"))
            store.insert_row(table, ctx, SampleRow("src/bar.py", "world"))

            assert store.row_counts["org.openrewrite.test.TestTable"] == 2
            assert "org.openrewrite.test.TestTable" in store.table_names



__path__ = __import__('pkgutil').extend_path(__path__, __name__)

from typing import Optional

from .auto_format import AutoFormat, AutoFormatVisitor
from .blank_lines import BlankLinesVisitor
from .minimum_viable_spacing import MinimumViableSpacingVisitor
from .normalize_format import NormalizeFormatVisitor
from .normalize_tabs_or_spaces import NormalizeTabsOrSpacesVisitor
from .spaces_visitor import SpacesVisitor
from .tabs_and_indents_visitor import TabsAndIndentsVisitor

from ...visitor import Cursor


def auto_format(tree, p, stop_after=None, cursor: Optional[Cursor] = None):
    try:
        return AutoFormatVisitor(stop_after).visit(tree, p, cursor)
    except (ValueError, AttributeError):
        return tree


def maybe_auto_format(before, after, p, stop_after=None, cursor: Optional[Cursor] = None):
    if before is not after:
        return auto_format(after, p, stop_after, cursor)
    return after


__all__ = [
    'AutoFormat',
    'AutoFormatVisitor',
    'BlankLinesVisitor',
    'MinimumViableSpacingVisitor',
    'NormalizeFormatVisitor',
    'NormalizeTabsOrSpacesVisitor',
    'SpacesVisitor',
    'TabsAndIndentsVisitor',
    'auto_format',
    'maybe_auto_format',
]



# Copyright 2025 the original author or authors.
# <p>
# Licensed under the Moderne Source Available License (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# <p>
# https://docs.moderne.io/licensing/moderne-source-available-license
# <p>
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Pattern matching comparator for comparing pattern and target ASTs."""

from __future__ import annotations

from typing import Dict, Optional, TYPE_CHECKING, cast

from rewrite.java import J
from rewrite.java import tree as j
from rewrite.python import tree as py
from .capture import Capture
from .placeholder import from_placeholder

if TYPE_CHECKING:
    from rewrite.visitor import Cursor


class PatternMatchingComparator:
    """
    Compares a pattern AST against a target AST, extracting captures.

    The comparator walks both trees in parallel, comparing node structures
    and types. When it encounters a placeholder identifier in the pattern,
    it captures the corresponding subtree from the target.

    Examples:
        comparator = PatternMatchingComparator(captures)
        result = comparator.match(pattern_tree, target_tree, cursor)
        if result:
            # result is a dict of capture_name -> captured_value
            x_value = result['x']
    """

    def __init__(
        self,
        captures: Dict[str, Capture],
        lenient_type_matching: bool = True
    ):
        """
        Initialize the comparator.

        Args:
            captures: Dict mapping capture names to Capture objects.
            lenient_type_matching: If True, ignore type differences when comparing.
        """
        self._captures = captures
        self._lenient_type_matching = lenient_type_matching
        self._captured: Dict[str, J] = {}
        self._debug = False

    def match(
        self,
        pattern: J,
        target: J,
        cursor: 'Cursor',
        *,
        debug: bool = False
    ) -> Optional[Dict[str, J]]:
        """
        Match pattern against target, returning captures if successful.

        Args:
            pattern: The pattern AST to match.
            target: The target AST to match against.
            cursor: Cursor at the target's position.
            debug: Enable debug logging.

        Returns:
            Dict of captured values if matched, None otherwise.
        """
        self._captured = {}
        self._debug = debug

        if self._compare(pattern, target, cursor):
            # Validate all required captures were matched
            for name, cap in self._captures.items():
                if name not in self._captured and not cap.variadic:
                    if self._debug:
                        print(f"Required capture '{name}' was not matched")
                    return None

            # Validate constraints
            for name, value in self._captured.items():
                cap = self._captures.get(name)
                if cap and cap.constraint:
                    if not cap.constraint(value):
                        if self._debug:
                            print(f"Capture '{name}' failed constraint")
                        return None

            return self._captured
        return None

    def _compare(
        self,
        pattern: Optional[J],
        target: Optional[J],
        cursor: 'Cursor'
    ) -> bool:
        """
        Recursively compare pattern and target nodes.

        Args:
            pattern: Pattern node (may contain placeholders).
            target: Target node to compare against.
            cursor: Current cursor position.

        Returns:
            True if nodes match.
        """
        # Both None is a match
        if pattern is None and target is None:
            return True

        # Only one None is not a match
        if pattern is None or target is None:
            return False

        # Check if pattern is a capture placeholder
        if isinstance(pattern, j.Identifier):
            capture_name = from_placeholder(pattern.simple_name)
            if capture_name is not None:
                return self._capture_node(capture_name, target)

        # Types must match (unless lenient)
        if type(pattern) != type(target):
            if self._debug:
                print(f"Type mismatch: {type(pattern).__name__} vs {type(target).__name__}")
            return False

        # Dispatch to type-specific comparison
        if isinstance(pattern, j.Identifier):
            return self._compare_identifier(pattern, cast(j.Identifier, target))
        elif isinstance(pattern, j.Literal):
            return self._compare_literal(pattern, cast(j.Literal, target))
        elif isinstance(pattern, j.Empty):
            # Two Empty sentinel nodes always match (used for absent values)
            return True
        elif isinstance(pattern, j.MethodInvocation):
            return self._compare_method_invocation(pattern, cast(j.MethodInvocation, target), cursor)
        elif isinstance(pattern, j.FieldAccess):
            return self._compare_field_access(pattern, cast(j.FieldAccess, target), cursor)
        elif isinstance(pattern, j.Binary):
            return self._compare_binary(pattern, cast(j.Binary, target), cursor)
        elif isinstance(pattern, j.Unary):
            return self._compare_unary(pattern, cast(j.Unary, target), cursor)
        elif isinstance(pattern, j.Assignment):
            return self._compare_assignment(pattern, cast(j.Assignment, target), cursor)
        elif isinstance(pattern, j.Parentheses):
            return self._compare_parentheses(pattern, cast(j.Parentheses, target), cursor)
        elif isinstance(pattern, j.Ternary):
            return self._compare_ternary(pattern, cast(j.Ternary, target), cursor)
        elif isinstance(pattern, j.Return):
            return self._compare_return(pattern, cast(j.Return, target), cursor)
        elif isinstance(pattern, j.ArrayAccess):
            return self._compare_array_access(pattern, cast(j.ArrayAccess, target), cursor)
        elif isinstance(pattern, py.ExpressionStatement):
            return self._compare_expression_statement(pattern, cast(py.ExpressionStatement, target), cursor)
        elif isinstance(pattern, py.Binary):
            return self._compare_python_binary(pattern, cast(py.Binary, target), cursor)
        elif isinstance(pattern, py.CollectionLiteral):
            return self._compare_collection_literal(pattern, cast(py.CollectionLiteral, target), cursor)
        elif isinstance(pattern, py.DictLiteral):
            return self._compare_dict_literal(pattern, cast(py.DictLiteral, target), cursor)
        else:
            # Default: unhandled node type — reject the match to prevent
            # false positives.  If a pattern uses a node type that reaches
            # this branch, a specific comparator method should be added.
            if self._debug:
                print(f"No specific comparison for {type(pattern).__name__}, rejecting match")
            return False

    def _capture_node(self, name: str, target: J) -> bool:
        """
        Capture a target node for a placeholder.

        Args:
            name: The capture name.
            target: The target node to capture.

        Returns:
            True if capture succeeded.
        """
        if name not in self._captures:
            if self._debug:
                print(f"Unknown capture: {name}")
            return False

        cap = self._captures[name]

        # Check if this capture already has a value
        if name in self._captured:
            # Must match the existing captured value
            existing = self._captured[name]
            # For now, require exact same node (by id)
            # TODO: implement semantic equality
            if self._debug:
                print(f"Capture '{name}' already has value, checking match")
            return existing.id == target.id

        # New capture
        self._captured[name] = target
        if self._debug:
            print(f"Captured '{name}': {type(target).__name__}")
        return True

    def _compare_identifier(self, pattern: j.Identifier, target: j.Identifier) -> bool:
        """Compare two identifiers."""
        return pattern.simple_name == target.simple_name

    def _compare_literal(self, pattern: j.Literal, target: j.Literal) -> bool:
        """Compare two literals.

        Uses a two-level comparison:
        1. Value types must match (prevents cross-type false positives like
           None vs b"" where both might serialize to the same representation).
        2. When both values are None (Ellipsis, None keyword, or literals with
           unicode escapes all store value=None), fall back to value_source
           comparison to distinguish them.
        """
        if type(pattern.value) != type(target.value):
            return False
        if pattern.value is None:
            return pattern.value_source == target.value_source
        return pattern.value == target.value

    def _compare_method_invocation(
        self,
        pattern: j.MethodInvocation,
        target: j.MethodInvocation,
        cursor: 'Cursor'
    ) -> bool:
        """Compare two method invocations."""
        # Compare select (receiver)
        if not self._compare(pattern.select, target.select, cursor):
            return False

        # Compare method name
        if not self._compare(pattern.name, target.name, cursor):
            return False

        # Compare arguments
        return self._compare_arguments(pattern.padding.arguments, target.padding.arguments, cursor)

    def _compare_arguments(
        self,
        pattern_args: Optional[j.JContainer],
        target_args: Optional[j.JContainer],
        cursor: 'Cursor'
    ) -> bool:
        """Compare argument containers."""
        if pattern_args is None and target_args is None:
            return True
        if pattern_args is None or target_args is None:
            return False

        pattern_elements = pattern_args.elements
        target_elements = target_args.elements

        # Check for variadic capture
        if len(pattern_elements) == 1:
            pattern_arg = pattern_elements[0]
            if isinstance(pattern_arg, j.Identifier):
                cap_name = from_placeholder(pattern_arg.simple_name)
                if cap_name and self._captures.get(cap_name, Capture(name=cap_name)).variadic:
                    # Variadic capture - capture all target arguments
                    # TODO: implement proper variadic handling
                    return True

        # Non-variadic: must have same number of arguments
        if len(pattern_elements) != len(target_elements):
            if self._debug:
                print(f"Argument count mismatch: {len(pattern_elements)} vs {len(target_elements)}")
            return False

        # Compare each argument
        for p_elem, t_elem in zip(pattern_elements, target_elements):
            if not self._compare(p_elem, t_elem, cursor):
                return False

        return True

    def _compare_field_access(
        self,
        pattern: j.FieldAccess,
        target: j.FieldAccess,
        cursor: 'Cursor'
    ) -> bool:
        """Compare two field accesses."""
        # Compare target (receiver)
        if not self._compare(pattern.target, target.target, cursor):
            return False

        # Compare name
        return self._compare(pattern.name, target.name, cursor)

    def _compare_binary(
        self,
        pattern: j.Binary,
        target: j.Binary,
        cursor: 'Cursor'
    ) -> bool:
        """Compare two Java binary expressions."""
        # Compare operator
        if pattern.operator != target.operator:
            return False

        # Compare operands
        if not self._compare(pattern.left, target.left, cursor):
            return False

        return self._compare(pattern.right, target.right, cursor)

    def _compare_python_binary(
        self,
        pattern: py.Binary,
        target: py.Binary,
        cursor: 'Cursor'
    ) -> bool:
        """Compare two Python binary expressions."""
        # Compare operator
        if pattern.operator != target.operator:
            return False

        # Compare operands
        if not self._compare(pattern.left, target.left, cursor):
            return False

        return self._compare(pattern.right, target.right, cursor)

    def _compare_unary(
        self,
        pattern: j.Unary,
        target: j.Unary,
        cursor: 'Cursor'
    ) -> bool:
        """Compare two unary expressions."""
        if pattern.operator != target.operator:
            return False

        return self._compare(pattern.expression, target.expression, cursor)

    def _compare_assignment(
        self,
        pattern: j.Assignment,
        target: j.Assignment,
        cursor: 'Cursor'
    ) -> bool:
        """Compare two assignments."""
        if not self._compare(pattern.variable, target.variable, cursor):
            return False

        return self._compare(pattern.assignment, target.assignment, cursor)

    def _compare_ternary(
        self,
        pattern: j.Ternary,
        target: j.Ternary,
        cursor: 'Cursor'
    ) -> bool:
        """Compare two ternary (conditional) expressions."""
        if not self._compare(pattern.condition, target.condition, cursor):
            return False

        if not self._compare(pattern.true_part, target.true_part, cursor):
            return False

        return self._compare(pattern.false_part, target.false_part, cursor)

    def _compare_array_access(
        self,
        pattern: j.ArrayAccess,
        target: j.ArrayAccess,
        cursor: 'Cursor'
    ) -> bool:
        """Compare two array/subscript accesses."""
        if not self._compare(pattern.indexed, target.indexed, cursor):
            return False

        return self._compare(pattern.dimension.index, target.dimension.index, cursor)

    def _compare_parentheses(
        self,
        pattern: j.Parentheses,
        target: j.Parentheses,
        cursor: 'Cursor'
    ) -> bool:
        """Compare two parenthesized expressions."""
        return self._compare(pattern.tree, target.tree, cursor)

    def _compare_return(
        self,
        pattern: j.Return,
        target: j.Return,
        cursor: 'Cursor'
    ) -> bool:
        """Compare two return statements."""
        return self._compare(pattern.expression, target.expression, cursor)

    def _compare_expression_statement(
        self,
        pattern: py.ExpressionStatement,
        target: py.ExpressionStatement,
        cursor: 'Cursor'
    ) -> bool:
        """Compare two Python expression statements."""
        return self._compare(pattern.expression, target.expression, cursor)

    def _compare_collection_literal(
        self,
        pattern: py.CollectionLiteral,
        target: py.CollectionLiteral,
        cursor: 'Cursor'
    ) -> bool:
        """Compare two Python collection literals."""
        # Compare kind (list, set, tuple)
        if pattern.kind != target.kind:
            return False

        # Compare elements
        pattern_elements = pattern.elements
        target_elements = target.elements

        if len(pattern_elements) != len(target_elements):
            return False

        for p_elem, t_elem in zip(pattern_elements, target_elements):
            if not self._compare(p_elem, t_elem, cursor):
                return False

        return True

    def _compare_dict_literal(
        self,
        pattern: py.DictLiteral,
        target: py.DictLiteral,
        cursor: 'Cursor'
    ) -> bool:
        """Compare two Python dict literals."""
        pattern_elements = pattern.elements
        target_elements = target.elements

        if len(pattern_elements) != len(target_elements):
            return False

        for p_elem, t_elem in zip(pattern_elements, target_elements):
            if not self._compare(p_elem, t_elem, cursor):
                return False

        return True



from __future__ import annotations

from dataclasses import replace, dataclass
from typing import TypeVar, Any, Optional, TYPE_CHECKING

from rewrite import TreeVisitor, Markers
from rewrite.java.tree import J
from rewrite.java import Comment

if TYPE_CHECKING:
    from .visitor import PythonVisitor

P = TypeVar('P')


class Py(J):
    def accept(self, v: TreeVisitor[Any, P], p: P) -> Optional[Any]:
        from .visitor import PythonVisitor
        return self.accept_python(v.adapt(Py, PythonVisitor), p)

    def accept_python(self, v: 'PythonVisitor[P]', p: P) -> Optional['J']:
        ...

    def is_acceptable(self, v: TreeVisitor[Any, P], p: P) -> bool:
        from .visitor import PythonVisitor
        return v.is_adaptable_to(PythonVisitor)


T = TypeVar('T')
J2 = TypeVar('J2', bound=J)


@dataclass(frozen=True)
class PyComment(Comment):
    _aligned_to_indent: bool

    @property
    def aligned_to_indent(self) -> bool:
        return self._aligned_to_indent

    def with_aligned_to_indent(self, aligned_to_indent: bool) -> Comment:
        return self if aligned_to_indent is self._aligned_to_indent else replace(self,
                                                                                 _aligned_to_indent=aligned_to_indent)

    @property
    def multiline(self) -> bool:
        return False

    # IMPORTANT: This explicit constructor aligns the parameter order with the Java side
    def __init__(self, _text: str, _suffix: str, _aligned_to_indent: bool, _markers: Markers) -> None:
        object.__setattr__(self, '_text', _text)
        object.__setattr__(self, '_suffix', _suffix)
        object.__setattr__(self, '_aligned_to_indent', _aligned_to_indent)
        object.__setattr__(self, '_markers', _markers)



from __future__ import annotations

import os
import threading
from abc import ABC, abstractmethod
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from typing import Optional, Any, TypeVar, cast, TYPE_CHECKING, Generic, ClassVar, Callable, Type
from uuid import UUID

from .markers import Markers
from .style import NamedStyles, Style
from .utils import replace_if_changed

if TYPE_CHECKING:
    from rewrite import TreeVisitor, ExecutionContext
    from .markers import Marker
    from .parser import ParserInput
    from .visitor import Cursor

P = TypeVar('P')


class Tree(ABC):
    @property
    @abstractmethod
    def id(self) -> UUID:
        ...

    @property
    @abstractmethod
    def markers(self) -> Markers:
        ...

    @abstractmethod
    def is_acceptable(self, v: TreeVisitor[Any, P], p: P) -> bool:
        ...

    def accept(self, v: TreeVisitor[Any, P], p: P) -> Optional[Any]:
        return v.default_value(self, p)

    def print(self, cursor: 'Cursor', capture: 'PrintOutputCapture[P]') -> str:
        printer = self.printer(cursor)
        printer.set_cursor(cursor)
        printer.visit(self, capture)
        return capture.get_out()

    def printer(self, cursor: 'Cursor') -> 'TreeVisitor[Any, PrintOutputCapture[P]]':
        return cursor.first_enclosing_or_throw(SourceFile).printer(cursor)

    def is_scope(self, tree: Optional[Tree]) -> bool:
        return tree is not None and tree.id == self.id

    def __eq__(self, other: object) -> bool:
        if self.__class__ == other.__class__:
            return self.id == cast(Tree, other).id
        return False

    def __hash__(self) -> int:
        return hash(self.id)

    def replace(self, **kwargs) -> 'Tree':
        """Replace fields on this tree node, returning self if nothing changed."""
        return replace_if_changed(self, **kwargs)


class PrinterFactory(ABC):
    _thread_local = threading.local()

    @classmethod
    def current(cls) -> Optional[PrinterFactory]:
        return getattr(PrinterFactory._thread_local, 'context', None)

    def set_current(self):
        PrinterFactory._thread_local.context = self

    @abstractmethod
    def create_printer(self, cursor: Cursor) -> TreeVisitor[Any, PrintOutputCapture[P]]:
        ...


S = TypeVar('S', bound=Style)

class SourceFile(Tree):
    @property
    @abstractmethod
    def charset_name(self) -> Optional[str]:
        ...

    @property
    @abstractmethod
    def source_path(self) -> Path:
        ...

    @property
    @abstractmethod
    def file_attributes(self) -> Optional[FileAttributes]:
        ...

    def print_all(self) -> str:
        from .visitor import Cursor
        return self.print(Cursor(None, Cursor.ROOT_VALUE), PrintOutputCapture(0))

    def print_equals_input(self, input: 'ParserInput', ctx: ExecutionContext) -> bool:
        printed = self.print_all()
        return printed == input.source().read()

    def get_style(self, style: Type[S]) -> Optional[S]:
        return NamedStyles.merge(style, self.markers.find_all(NamedStyles))


@dataclass(frozen=True)
class FileAttributes:
    creation_time: Optional[datetime]
    last_modified_time: Optional[datetime]
    last_access_time: Optional[datetime]
    is_readable: bool
    is_writable: bool
    is_executable: bool
    size: int

    @staticmethod
    def from_path(path: Path) -> Optional[FileAttributes]:
        if path.exists():
            try:
                # Get file stats
                stat = path.stat()
                creation_time = datetime.fromtimestamp(stat.st_ctime)
                last_modified_time = datetime.fromtimestamp(stat.st_mtime)
                last_access_time = datetime.fromtimestamp(stat.st_atime)

                is_readable = os.access(path, os.R_OK)
                is_writable = os.access(path, os.W_OK)
                is_executable = os.access(path, os.X_OK)
                size = stat.st_size

                return FileAttributes(creation_time, last_access_time, last_modified_time, is_readable, is_writable,
                                      is_executable, size)
            except OSError:
                pass
        return None


@dataclass(frozen=True)
class Checksum:
    algorithm: str
    value: bytes


class PrintOutputCapture(Generic[P]):
    @dataclass
    class MarkerPrinter(ABC):
        DEFAULT: ClassVar[Optional['PrintOutputCapture.MarkerPrinter']] = None  # Set at module load
        SEARCH_MARKERS_ONLY: ClassVar[Optional['PrintOutputCapture.MarkerPrinter']] = None  # Set at module load
        VERBOSE: ClassVar[Optional['PrintOutputCapture.MarkerPrinter']] = None  # Set at module load
        FENCED: ClassVar[Optional['PrintOutputCapture.MarkerPrinter']] = None  # Set at module load
        SANITIZED: ClassVar[Optional['PrintOutputCapture.MarkerPrinter']] = None  # Set at module load

        def before_syntax(self, marker: 'Marker', cursor: 'Cursor', comment_wrapper: Callable[[str], str]) -> str:
            return ""

        def before_prefix(self, marker: 'Marker', cursor: 'Cursor', comment_wrapper: Callable[[str], str]) -> str:
            return ""

        def after_syntax(self, marker: 'Marker', cursor: 'Cursor', comment_wrapper: Callable[[str], str]) -> str:
            return ""

    def __init__(self, p: P, marker_printer: Optional['PrintOutputCapture.MarkerPrinter'] = None):
        self._context = p
        self._marker_printer: PrintOutputCapture.MarkerPrinter = marker_printer or PrintOutputCapture.MarkerPrinter.DEFAULT  # ty: ignore[invalid-assignment]  # DEFAULT always set at module load
        self._out: list[str] = []

    def get_out(self) -> str:
        return ''.join(self._out)

    def append(self, text: Optional[str] = None) -> 'PrintOutputCapture[P]':
        if text and len(text) > 0:
            self._out.append(text)
        return self

    def append_char(self, c: str) -> 'PrintOutputCapture[P]':
        if len(c) == 1:
            self._out.append(c)
        return self

    def clone(self) -> 'PrintOutputCapture[P]':
        return PrintOutputCapture(self._context, self._marker_printer)

    @property
    def marker_printer(self) -> MarkerPrinter:
        return self._marker_printer

    def get_marker_printer(self) -> MarkerPrinter:
        """Return the marker printer (alias for marker_printer property)."""
        return self._marker_printer


@dataclass
class _DefaultMarkerPrinter(PrintOutputCapture.MarkerPrinter):
    def before_syntax(self, marker: 'Marker', cursor: 'Cursor', comment_wrapper: Callable[[str], str]) -> str:
        return marker.print(cursor, comment_wrapper, False)


@dataclass
class _SearchMarkersOnlyPrinter(PrintOutputCapture.MarkerPrinter):
    """Only prints SearchResult markers, ignores all others."""
    def before_syntax(self, marker: 'Marker', cursor: 'Cursor', comment_wrapper: Callable[[str], str]) -> str:
        from .markers import SearchResult
        if isinstance(marker, SearchResult):
            return marker.print(cursor, comment_wrapper, False)
        return ""


@dataclass
class _VerboseMarkerPrinter(PrintOutputCapture.MarkerPrinter):
    """Prints all markers with verbose=True (shows detail instead of message)."""
    def before_syntax(self, marker: 'Marker', cursor: 'Cursor', comment_wrapper: Callable[[str], str]) -> str:
        return marker.print(cursor, comment_wrapper, True)


@dataclass
class _FencedMarkerPrinter(PrintOutputCapture.MarkerPrinter):
    """Prints SearchResult and Markup markers with fenced {{uuid}} format."""
    def before_syntax(self, marker: 'Marker', cursor: 'Cursor', comment_wrapper: Callable[[str], str]) -> str:
        from .markers import SearchResult, Markup  # ty: ignore[unresolved-import]
        if isinstance(marker, (SearchResult, Markup)):
            return "{{" + str(marker.id) + "}}"
        return ""

    def after_syntax(self, marker: 'Marker', cursor: 'Cursor', comment_wrapper: Callable[[str], str]) -> str:
        from .markers import SearchResult, Markup  # ty: ignore[unresolved-import]
        if isinstance(marker, (SearchResult, Markup)):
            return "{{" + str(marker.id) + "}}"
        return ""


@dataclass
class _SanitizedMarkerPrinter(PrintOutputCapture.MarkerPrinter):
    """Prints nothing - produces clean output without any marker annotations."""
    pass


PrintOutputCapture.MarkerPrinter.DEFAULT = _DefaultMarkerPrinter()
PrintOutputCapture.MarkerPrinter.SEARCH_MARKERS_ONLY = _SearchMarkersOnlyPrinter()
PrintOutputCapture.MarkerPrinter.VERBOSE = _VerboseMarkerPrinter()
PrintOutputCapture.MarkerPrinter.FENCED = _FencedMarkerPrinter()
PrintOutputCapture.MarkerPrinter.SANITIZED = _SanitizedMarkerPrinter()
